from collections import OrderedDict, defaultdict

import six
from six.moves import configparser



class LineTrackingMixin(object):
    def __init__(self):
        # We'll call constructors explicitly to ensure that this happens after
        # the main base class.

        self._option_lines = defaultdict(dict)

        # State:
        # lineno = None, reading_file = None    => not reading
        # lineno = int, reading_file = str    => reading lines
        # lineno = None, reading_file = str    => finalizing parsing
        self._curr_lineno = None
        self._curr_filename = None

        # Use self._dict as the base class in case the user used something
        # custom. Hold on to a reference though, because it may change

        dict_base = self._dict
        class OptionWrapper(self._dict):
            def __init__(inner, *args, **kwargs):
                inner.sectname = None
                dict_base.__init__(inner, *args, **kwargs)

            def __setitem__(inner, key, value):
                sectname = inner.sectname
                if sectname is not None:
                    if self._curr_lineno is not None:
                        location = self._curr_filename, self._curr_lineno
                        self._option_lines[sectname][key] = location
                    elif self._curr_filename is None and sectname is not None:
                        self._option_lines[sectname].pop(key, None)
                dict_base.__setitem__(inner, key, value)

            def __delitem__(inner, key):
                dict_base.__delitem__(inner, key)
                if inner.sectname is not None:
                    self._option_lines[inner.sectname].pop(key, None)

            def pop(inner, key, *args, **kwargs):
                val = dict_base.pop(inner, key, *args, **kwargs)
                if inner.sectname is not None:
                    self._option_lines[inner.sectname].pop(key, None)
                return val


        class SectionWrapper(self._dict):
            def __setitem__(inner, key, value):
                value.sectname = key
                dict_base.__setitem__(inner, key, value)

            def __delitem__(inner, key):
                dict_base.__delitem__(inner, key)
                del self._option_lines[key]

            def pop(inner, key, *args, **kwargs):
                val = dict_base.pop(inner, key, *args, **kwargs)
                self._option_lines.pop(key, None)
                return val

        class FpWrapper(object):
            def __init__(inner, fp):
                inner.fp = fp
                self._curr_lineno = None

            def readline(inner):
                line = inner.fp.readline()
                if not line:
                    self._curr_lineno = None
                elif self._curr_lineno is None:
                    self._curr_lineno = 1
                else:
                    self._curr_lineno += 1
                return line
            
            def __iter__(inner):
                for line in inner.fp:
                    if self._curr_lineno is None:
                        self._curr_lineno = 1
                    else:
                        self._curr_lineno += 1
                    yield line
                self._curr_lineno = None
        self._fp_wrapper = FpWrapper

        self._dict = OptionWrapper
        self._sections = SectionWrapper()
        self._defaults = OptionWrapper(self._defaults)
        self._defaults.sectname = configparser.DEFAULTSECT

    def _read(self, fp, fpname):
        self._curr_filename = fpname
        fp = self._fp_wrapper(fp)
        try:
            super(LineTrackingMixin, self)._read(fp, fpname)
        finally:
            self._curr_filename = None
            self._curr_lineno = None

    def get_location(self, section, option):
        if not self.has_section(section):
            raise configparser.NoSectionError(section)
        elif not self.has_option(section, option):
            raise configparser.NoOptionError(option, section)
        option = self.optionxform(option)
        loc = self._option_lines[section].get(option)
        if loc is None and option in self._defaults:
            return self._option_lines[configparser.DEFAULTSECT].get(option)
        return loc

    def get_line(self, section, option):
        loc = self.get_location(section, option)
        if loc is not None:
            return loc[1]
        return None

    def get_filename(self, section, option):
        loc = self.get_location(section, option)
        if loc is not None:
            return loc[0]
        return None

    def set(self, section, option, value, *args, **kwargs):
        new_location = kwargs.pop('location', None)
        if new_location not in ('preserve', None):
            try:
                filename, lineno = new_location
            except ValueError:
                err = ValueError("location must be (filename, lineno), None, "
                                 "or 'preserve'")
                six.raise_from(err, None)

        option_xform = self.optionxform(option)
        cur_location = self._option_lines[section].get(option_xform)
        super(LineTrackingMixin, self).set(section, option, value,
                *args, **kwargs)
        if new_location == 'preserve':
            if cur_location is not None:
                self._option_lines[section][option_xform] = cur_location
        elif new_location is not None:
            self._option_lines[section][option_xform] = new_location

    def set_location(self, section, option, location):
        if not self.has_section(section):
            raise configparser.NoSectionError(section)
        elif not self.has_option(section, option):
            raise configparser.NoOptionError(option, section)

        if location is not None:
            try:
                filename, lineno = location
            except ValueError:
                err = ValueError("location must be (filename, lineno) or None")
                six.raise_from(err, None)
        option = self.optionxform(option)
        self._option_lines[section][option] = location

class RawConfigParser(LineTrackingMixin, configparser.RawConfigParser):
    def __init__(self, *args, **kwargs):
        configparser.RawConfigParser.__init__(self, *args, **kwargs)
        LineTrackingMixin.__init__(self)

class ConfigParser(LineTrackingMixin, configparser.ConfigParser):
    def __init__(self, *args, **kwargs):
        configparser.ConfigParser.__init__(self, *args, **kwargs)
        LineTrackingMixin.__init__(self)

class SafeConfigParser(LineTrackingMixin, configparser.SafeConfigParser):
    def __init__(self, *args, **kwargs):
        configparser.SafeConfigParser.__init__(self, *args, **kwargs)
        LineTrackingMixin.__init__(self)

